__title__ = 'mebo'
__description__ = 'Simple python interface to control the mebo toy robot'
__url__ = 'https://github.com/crlane/python-mebo'
__version__ = '0.1.0.b1'
__author__ = 'Cameron Lane'
__author_email__ = 'crlane@adamanteus.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2018 Cameron Lane'
